import asyncio
import inspect
import json
from multiprocessing import Process, Pipe
from multiprocessing.connection import Connection
from types import FunctionType


class ProcessConnector:
    def __init__(self, function):
        self.parent_con, self.child_con = None, None
        self.pid = 0
        self.proc = None
        self.raw_func = function
        self.is_async = inspect.iscoroutinefunction(function)
        self.result = None

    def kill(self):
        self.proc.kill()

    def execute_in_process(self, conn: Connection):
        try:
            input_params = conn.recv()
            args = input_params.get("args", [])
            kwargs = input_params.get("kwargs", {})
            result = self.raw_func(*args, **kwargs)
            try:
                json.dumps(result)
            except Exception as e:
                print(e)
                result = {}
            conn.send(result)
        except Exception as error:
            print(error)
            raise error
        finally:
            conn.close()

    async def execute_in_process_event_loop(self, conn: Connection):
        try:
            input_params = conn.recv()
            args = input_params.get("args", [])
            kwargs = input_params.get("kwargs", {})
            result = self.raw_func(*args, **kwargs)
            try:
                json.dumps(result)
            except Exception as e:
                print(e)
                result = {}
            conn.send(result)
        except Exception as error:
            print(error)
            raise error
        finally:
            conn.close()

    def create_and_run(self, *args, **kwargs):
        self.parent_con, self.child_con = Pipe()
        if self.is_async:
            self.proc = Process(target=asyncio.run(self.execute_in_process_event_loop(self.child_con)))
        else:
            self.proc = Process(target=lambda: self.execute_in_process(self.child_con))
        self.proc.start()
        self.parent_con.send({"args": args, "kwargs": kwargs})
        return self.parent_con

    @property
    def ret(self):
        if self.parent_con and not self.parent_con.closed and self.child_con.closed:
            self.result = self.parent_con.recv()
            self.parent_con.close()
        return self.result

    def wait(self, timeout):
        self.proc.join(timeout)

    def receive(self):
        return self.parent_con.recv()

    @staticmethod
    def create_process(function: FunctionType):
        assert isinstance(function, FunctionType), "only accept function for process"
        def run_in_different_proc(*args, **kwargs):
            pc = ProcessConnector(function)
            pc.create_and_run(*args, **kwargs)
            return pc
        return run_in_different_proc



from .worker import *

### SHORTCUT ###
run_as_worker = ThreadWorkerManager.run_as_worker
abort_worker = ThreadWorkerManager.abort_worker
abort_all_worker = ThreadWorkerManager.abort_all_worker
abort_thread = ThreadWorkerManager.abort_thread
abort_all_thread = ThreadWorkerManager.abort_all_thread
enableKeyboardInterrupt = ThreadWorkerManager.enableKeyboardInterrupt
disableKeyboardInterrupt = ThreadWorkerManager.disableKeyboardInterrupt


@overload
def worker(function: FunctionType) -> ThreadedFunction: ...


@overload
def worker(
    name: Optional[str] = "",
    on_abort: Optional[FunctionType] = None,
    keyboard_interrupt: Optional[bool] = True
) -> ThreadedFunction: ...


@overload
def worker(
    on_abort: Optional[FunctionType] = None,
    keyboard_interrupt: Optional[bool] = True
) -> ThreadedFunction: ...


def worker(
    *args,
    name: Optional[str] = "",
    on_abort: Optional[FunctionType] = None,
    keyboard_interrupt: Optional[bool] = True,
    timeout: float = 0,
    **kargs
):
    """
    A worker decorator.
    Turn a main-thread function into a background-thread function automatically

    Usage Example:
    - @worker
    - @worker("cool")
    - @worker(name="looping backapp", keyboard_interrupt=True)
    - @worker(keyboard_interrupt=True, on_abort: lambda: print("its over"))
    """
    if args:
        if type(args[0]) == FunctionType:
            assert not inspect.iscoroutinefunction(args[0]), "please use `async_worker` instead for coroutine function"
            def register(*dargs,**dkargs):
                w = ThreadWorker(lambda: args[0](*dargs,**dkargs), name, on_abort, keyboard_interrupt)
                w.work()
                return w
            return register
        elif type(args[0]) == str:
            name = args[0]
            return ThreadWorkerManager.create_worker(name, *args[1:], on_abort=on_abort, keyboard_interrupt=keyboard_interrupt, **kargs)
    else:
        return ThreadWorkerManager.create_worker(name, on_abort=on_abort, keyboard_interrupt=keyboard_interrupt, **kargs)


@overload
def async_worker(function: FunctionType) -> AsyncThreadedFunction: ...

@overload
def async_worker(
    name: Optional[str] = "",
    on_abort: Optional[FunctionType] = None,
    keyboard_interrupt: Optional[bool] = True
) -> AsyncThreadedFunction: ...

@overload
def async_worker(
    on_abort: Optional[FunctionType] = None,
    keyboard_interrupt: Optional[bool] = True
) -> AsyncThreadedFunction: ...

def async_worker(
    *args,
    name: Optional[str] = "",
    on_abort: Optional[FunctionType] = None,
    keyboard_interrupt: Optional[bool] = True,
    **kargs
):
    """
    async_worker(function) -> threaded-corotine-function

    An asynchronous worker decorator.
    Turn a main-thread coroutine function into a background-thread coroutine function automatically

    Usage Example:
    - @async_worker
    - @async_worker("cool")
    - @async_worker(name="looping backapp", keyboard_interrupt=True)
    - @async_worker(keyboard_interrupt=True, on_abort: lambda: print("its over"))
    """
    if args:
        if type(args[0]) == FunctionType:
            assert inspect.iscoroutinefunction(args[0]), "please use `worker` instead for non-coroutine function"
            async def register(*dargs,**dkargs):
                async def worker_func():
                    return await args[0](*dargs,**dkargs)
                w = ThreadWorker(worker_func, name, on_abort, keyboard_interrupt)
                w.work()
                return w
            return register
        elif type(args[0]) == str:
            name = args[0]
            return ThreadWorkerManager.create_async_worker(name, *args[1:], on_abort=on_abort, keyboard_interrupt=keyboard_interrupt, **kargs)
    else:
        return ThreadWorkerManager.create_async_worker(name, on_abort=on_abort, keyboard_interrupt=keyboard_interrupt, **kargs)


from .process import *

def process(function: FunctionType):
    """
    Create a process worker. This function will run your function in a separate GIL
    """
    return ProcessConnector(function)